=== modified file 'cronscripts/librarian-gc.py'
--- cronscripts/librarian-gc.py	2005-11-11 15:24:27 +0000
+++ cronscripts/librarian-gc.py	2005-11-15 11:20:46 +0000
@@ -13,11 +13,7 @@
 from optparse import OptionParser
 
 from canonical.launchpad.scripts import logger_options, logger
-
-def merge_duplicates(ztm):
-    """Merge duplicate LibraryFileContent rows, and remove unlinked files
-    from disk
-    """
+from canonical.librarian import librariangc
 
 
 def main():
@@ -27,6 +23,7 @@
     (options, args) = parser.parse_args()
 
     log = logger(options)
+    librariangc.log = log
 
 
 if __name__ == '__main__':

=== added file 'database/schema/patch-25-57-0.sql'
--- database/schema/patch-25-57-0.sql	1970-01-01 00:00:00 +0000
+++ database/schema/patch-25-57-0.sql	2005-11-15 11:20:46 +0000
@@ -0,0 +1,8 @@
+SET client_min_messages=ERROR;
+
+DROP INDEX idx_libraryfilecontent_sha1;
+CREATE INDEX libraryfilecontent_sha1_filesize_idx
+    ON LibraryFileContent(sha1, filesize);
+
+INSERT INTO LaunchpadDatabaseRevision VALUES (25, 57, 0);
+

=== added file 'lib/canonical/librarian/librariangc.py'
--- lib/canonical/librarian/librariangc.py	1970-01-01 00:00:00 +0000
+++ lib/canonical/librarian/librariangc.py	2005-11-15 11:20:46 +0000
@@ -0,0 +1,82 @@
+# Copyright 2004-2005 Canonical Ltd.  All rights reserved.
+"""Librarian garbage collection routines"""
+
+__metaclass__ = type
+
+from canonical.database.sqlbase import cursor
+
+log = None
+
+def merge_duplicates(ztm):
+    """Merge duplicate LibraryFileContent rows"""
+
+    while True:
+        ztm.begin()
+
+        cur = cursor()
+        cur.execute("""
+            SELECT sha1, filesize
+            FROM LibraryFileContent
+            GROUP BY sha1, filesize
+            HAVING COUNT(*) > 1
+            LIMIT 1
+            """)
+        row = cur.fetchone()
+
+        if row is None:
+            # No more duplicates, so exit loop
+            break
+
+        sha1, filesize = cur.fetchone()
+
+        # Get a list of our dupes, making sure that the first in the
+        # list is not deleted if possible.
+        cur.execute("""
+            SELECT id
+            FROM LibraryFileContent
+            WHERE sha1=%(sha1) AND filesize=%(filesize)s
+            ORDER BY deleted, datecreated
+            """, vars())
+        dupes = [row[0] for row in cur.fetchall()]
+
+        log.info(
+                "Found duplicate LibraryFileContents %s",
+                ' '.join(dupes)
+                )
+
+        # Update all the LibraryFileAlias entries to point to a single
+        # LibraryFileContent
+        prime_id = dupes[0]
+        other_ids = ', '.join(dupes[1:])
+        log.debug(
+                "Making LibraryFileAliases referencing %s reference %s instead",
+                other_ids, prime_id
+                )
+        cur.execute("""
+            UPDATE LibraryFileAlias SET content=%(prime_id)s
+            WHERE content in (%(dupe_ids)s)
+            """)
+
+        log.debug("Committing")
+        ztm.commit()
+
+
+def delete_unreferenced_content(ztm):
+    """Delete LibraryFileContent entries and their disk files that are
+    not referenced by any LibraryFileAlias entries.
+    """
+    while True:
+        cur.execute("""
+            SELECT LibraryFileContent.id, deleted
+            FROM LibraryFileContent
+            LEFT OUTER JOIN LibraryFileAlias
+                ON LibraryFileContent.id = LibraryFileAlias.content
+            WHERE LibraryFileAlias.content IS NULL
+            LIMIT 1
+            """)
+
+        # Delete old LibraryFileContent entries
+        log.debug("Deleting LibraryFileContents %s", dupe_ids)
+        cur.execute("""
+            DELETE FROM LibraryFileContent WHERE id IN (%(dupe_ids)s)
+            """, vars())

